package highway

  1. Overview
  2. Docs

highway

API

Library highway

  • Highway Highway is a simple, framework-agnostic HTTP router. Its goal is to provide a highly abstract API based on the concepts of response and request so that it can, broadly speaking, be (probably) adapted to any framework in the OCaml ecosystem.