package highway

  1. Overview
  2. Docs

Module Highway.ArgsSource

Describes a heterogeneous list that uses arrow notation to characterize the various elements of the list.

The purpose of the list is solely to define route arguments. Its set of operations is therefore limited.

Types

Sourcetype _ t =
  1. | [] : Void.t t
  2. | :: : 'a * 'b t -> ('a -> 'b) t

Describes a heterogeneous list.