package highway

  1. Overview
  2. Docs

Module Highway.HandlerSource

A handler describes a function that takes a response as an argument and returns a request. Since Highway is generic, the library makes no assumptions about the type of the request or response; this is left to specialized frameworks.

Types

Sourcetype ('request, 'response) t = 'request -> 'response

A type describing a request handler.