package highway

  1. Overview
  2. Docs

Module Highway.Sigs

Set of Reusable Interfaces.

module type AS_PARAM = sig ... end

If a module implements this interface for a type t, it can be used to quickly describe a set of query parameters.