You can search for identifiers within the package.
in-package search v0.2.0
highway
Highway.Sigs
Set of Reusable Interfaces.
module type AS_PARAM = sig ... end
If a module implements this interface for a type t, it can be used to quickly describe a set of query parameters.
t